""" Package has implementation of ProGAN (progressive growing of GANs)
    as an extension of PyTorch Module
"""
import pro_gan_pytorch.PRO_GAN
import pro_gan_pytorch.CustomLayers
import pro_gan_pytorch.Losses
